# sparksdr2dxc.py (2024-05-01) by AndyZ K1RA https://k1ra.us/
#
#   Usage: sparksdr2dxc.py -m MYCALL [-c cache_expire_mins] [-s sparksdr_ip] [-t telnet_port] [-d] [-v] [-h]
#
#   DXcluster like telnet server providing spots from SparkSDR https://sparksdr.com/ Websocket data
#   for FT8, FT4, WSPR, FST4, FST4W, JS8 decodes across multiple bands.
#
# use "pip install -r requirements.txt" to install python dependencies and libraries needed by this app

# pyinstaller -n sparksdr2dxc.exe --onefile .\main.py

import json
import os
import re
import time
import threading
import sys
import getopt
import psutil as psutil

from telnetserver import TelnetServer
from multiprocessing import Process, Queue
from websocket import WebSocketApp
from threading import Thread
from queue import Queue
from functools import partial

version = "2.0.10"


class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time.time()
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time.time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


class Websocket_Client:
    def __init__(self, wsurl: str):
        self.responseQ = Queue()  # our message response Q

        # add the response Qs on by wrapping the callbacks in functools partials
        self.websocket = WebSocketApp(wsurl,
                                      on_open=partial(self.on_open, responseQ=self.responseQ),
                                      on_message=partial(self.on_message, responseQ=self.responseQ),
                                      on_error=partial(self.on_error, responseQ=self.responseQ),
                                      on_close=partial(self.on_close, responseQ=self.responseQ))

        # start run_forever as thread. request keep alive pings every 10 seconds with 5 second timeouts
        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    def on_open(self, websocket, responseQ):
        if debug:
            print("Opening Connection")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")
        time.sleep(1)
        websocket.send("{\"cmd\":\"subscribeToSpots\",\"Enable\":true}\n\n")

    def on_error(self, websocket, error, responseQ):
        if debug:
            print("Connection Error:", error)

    def on_message(self, websocket, message, responseQ):
        if debug:
            print("Message Received:", message)
        responseQ.put(message)

    def on_close(self, websocket, status, message, responseQ):
        if debug:
            print("Closing Connection:")
        time.sleep(2)

        self.controller = Thread(target=self.websocket.run_forever, args=(None, None, 10, 5))
        self.controller.start()

    def stop(self):
        self.websocket.keep_running = False
        self.websocket.close()


def parse_message(json_message):
    global debug
    global calls_db
    global cache_expire

    msg = ""

    date = json_message["time"]

    # parse time
    m = re.search('(\d{2}):(\d{2}):(\d{2})', date)
    hh = m.group(1)
    mm = m.group(2)
    ss = m.group(3)
    hhmmss = f"{hh}{mm}{ss}"

    call = json_message["call"]
    mode = json_message["mode"]

    freq = json_message["tunedfrequency"]
    freq = freq / 1000.

    tone = int(json_message["frequency"]) - int(json_message["tunedfrequency"])
    dist = json_message["distance"]
    snr = str(json_message["snr"])
    loc = json_message["locator"]

    # parse grid
    if loc == "":
        grid = "    "
    else:
        m = re.search('([A-Z][A-Z]\d\d)', loc)
        grid = m.group(1)

    dt = float(json_message["dt"])
    # valid = json_message["valid"]
    valid = ""
    jtmsg = json_message["msg"]

    jtmsg = re.sub('[\r\n]', '', jtmsg)

    called = ""

    # handle FTn messages
    if re.search('FT\d', mode) is not None:
        # determine if station is calling CQ or responding (DE) to another station
        cqde = jtmsg

        if re.search('CQ', cqde):
            cqde = "CQ"
        else:
            cqde = "DE"

        # look for different message structures and extract call and grid
        # CQ CALL
        if (m := re.search('^CQ\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)

        # CQ CALL LLnn
        elif (m := re.search('^CQ\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # [NA, DX, SD, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})$', jtmsg)) is not None:
            call = m.group(1)
            grid = ""

        # CQ [NA, DX, VT, TEST, xxx, etc] CALL LLnn
        elif (m := re.search('^CQ\s[A-Z]{2,}\s([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d$)', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 [R][-+]nn
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*[\-+][0-9]{2}$',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 RRR
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sRRR$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 GRID
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\s([A-Z][A-Z]\d\d$)',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL GRID
        elif (m := re.search('^\s*([A-Z\d\/]{3,})\s([A-Z][A-Z]\d\d)$', jtmsg)) is not None:
            call = m.group(1)
            grid = m.group(2)

        # CALL1 CALL2 RR73 or 73
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s\<?([A-Z\d\/]{3,})\>?\sR*73$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # SP9VEY RR73; IK6CLX <TY0RU> -08
        elif (
                m := re.search('^[A-Z\d\/]{3,}\sRR73;\s([A-Z\d\/]{3,})\s\<(\.\.\.|[A-Z\d\/]{3,})\>\s',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""

        # CALL1 CALL2 R GRID
        elif (
                m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\sR\s([A-Z][A-Z]\d\d$)',
                               jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = m.group(3)

        # CALL1 CALL2 RST nnnn
        elif (m := re.search('^\<?(\.\.\.|[A-Z\d\/]{3,})\>?\s([A-Z\d\/]{3,})\s\d{3}\s\d{4}$', jtmsg)) is not None:
            called = m.group(1)
            call = m.group(2)
            grid = ""
        # no match
        else:
            if debug:
                if call != "":
                    print(f"<<<--- NO MATCH --->>> {call}, {jtmsg}")
            return False, hhmmss, freq, mode, call, grid, tone, snr, dt, "", "", jtmsg

        if re.search('RR73', grid) is not None:
            grid = ""

        # form the comment msg for the DXcluster line for FTn
        msg = "{:>3} dB {:^4} {:>4} Hz {:^2} ".format(snr, grid, tone, cqde)

    if mode == "WSPR" or mode == "FST4W":
        # WSPR
        # watt = json_message["watts"]
        dbm = json_message["power"]

        # form the comment msg for the DXcluster line for WSPR
        msg = "{:>3} dB {:^4}   {:^2} dBm".format(snr, grid, dbm)
    else:
        watt = ""
        dbm = ""

    # handle JS8 messages
    if re.search('JS8', mode) is not None:
        # form the comment msg for the DXcluster line for FTn
        msg = "{:>3} dB {:^4} {:>4} Hz {:^2} ".format(snr, grid, tone, "DE")

    call = re.sub('[\<\>\.]+', '', call)

    # perform some sanity checks on calls return False if they're bogus
    # perform some sanity checks on calls, check for bogus combos and return False if necessary
    if (call == "" or re.search('RR73', call) is not None) or (re.search('\d[A-Z]', call) is None) or \
            (re.search('^\d{2;}[A-Z]{2,}', call) is not None) or (re.search('^[0Q]', call) is not None) or \
            (re.search('^[0Q]', call) is not None) or (re.search('^\d.\d.*\/[PQR]$', call) is not None) or \
            (re.search('^\d{2;}[A-Z]{2,}', called) is not None) or (re.search('_|\/.*\/]', call) is not None) or \
            (re.search('\d[A-Z]+\d[A-Z]+\d', call) is not None) or \
            ((re.search('\/[PQR]$', call) is not None) and (re.search('\/[PQR]$', called) is not None)) or \
            ((re.search('K1ABC', called) is not None) and (re.search('\/[PQR]$', call) is not None)):
        return False, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, msg, jtmsg

    # create hash of the callsign and frequency spotted
    callfreq = f"{call}{freq}"

    # get current time
    nowtime = int(time.time())

    # test if call was seen in last N expire_cache minutes
    if callfreq in calls_db:
        # call has been seen
        spottime = calls_db[callfreq]

        # test if call passed expiration limit
        if spottime <= nowtime - (cache_expire * 60):
            # yes, reset expire time and return decode
            calls_db[callfreq] = nowtime
        else:
            # no, its fresh and recently seen
            valid = "FRESH"
    else:
        # call not seen yet, save in database and return decode
        calls_db[callfreq] = nowtime

    return valid, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, msg, jtmsg


def clean_calls_db():
    global debug
    global calls_db
    global cache_expire

    if debug:
        print("Cleaning cached calls database...")

    # get current time in seconds
    nowtime = int(time.time())

    db_len = len(calls_db)

    # iterate through all call+freq hashes in database and drop any older than cache_expire timer (minutes)
    db = {callfreq: s_time for callfreq, s_time in calls_db.items() if s_time > nowtime - (cache_expire * 60)}

    if debug:
        calls_db = db
        new_len = len(db)
        print(f"{db_len - new_len} items removed from database")
        print()


# sparksdr2dxc.py SparkSDR_IP [cache timeout mins]
if __name__ == '__main__':
    debug = False
    cache_expire = 5
    sparksdr_ip = "127.0.0.1"
    sparksdr_port = 4649
    telnet_port = 7373
    mycall = ""

    try:
        opts, args = getopt.getopt(sys.argv[1:], "m:c:s:t:dhv")
    except getopt.GetoptError:
        print("sparksdr2dxc.py -m MYCALL [-c cache_expire_mins] [-s sparksdr_ip] [-t telnet_port] [-d] [-v] [-h]")
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print("Usage: sparksdr2dxc.py -m MYCALL [-c cache_expire] [-s sparksdr_ip] [-t telnet_port] [-d] [-v] [-h]")
            print("-m MYCALL (mandatory, ex: -m K1RA)")
            print("-c cache_expire (optional, minutes, ex: -c 5) [default=5]")
            print("-s sparksdr_ip (optional, ip address, ex: -s 192.168.1.234) [default=127.0.0.1]")
            print("-t telnet_port (optional, tcp port, ex: -t 2323) [default=7373]")
            print("-d (turns on debug messages)")
            print("-v (prints version)")
            print("-h (prints this help)")
            sys.exit()
        elif opt in "-c":
            cache_expire = arg
        elif opt in "-d":
            debug = True
        elif opt in "-m":
            mycall = arg
        elif opt in "-s":
            sparksdr_ip = arg
        elif opt in "-t":
            telnet_port = arg
        elif opt in "-v":
            print(f"sparkdxc2dxc version {version} by AndyZ K1RA")
            sys.exit()

    if mycall == "":
        print("sparksdr2dxc.py -m MYCALL [-c cache_expire_mins] [-s sparksdr_ip] [-t telnet_port] [-d] [-v] [-h]")
        sys.exit(2)

    print(f"sparkdxc2dxc version {version} by K1RA")

    url = f"ws://{sparksdr_ip}:{sparksdr_port}/Spark"

    # call+freq dictionary of received calls
    calls_db = {}

    # make the websocket client
    wsclient = Websocket_Client(url)

    # make the telnet server
    tserver = TelnetServer(port=int(telnet_port))

    # array list of incoming telnet clients
    tclients = []

    pid = os.getpid()

    clean_timer = int(time.time())

    try:
        # do forever
        while True:
            tserver.update()

            # For each newly connected client
            for new_client in tserver.get_new_clients():
                # Add them to the client list
                tclients.append(new_client)
                # Send a welcome message
                tserver.send_message(new_client, "login: ")
                tserver.send_message(new_client, "Welcome to SparkSDR2DXC version {}, you are client {}.".
                                     format(version, new_client))
                tserver.send_message(new_client, "CLIENT de SparkSDR2DXC>")
                calls_db = {}

            # For each client that has recently disconnected
            for disconnected_client in tserver.get_disconnected_clients():
                if disconnected_client not in tclients:
                    continue

                # Remove him from the clients list
                tclients.remove(disconnected_client)

                # Send every client a message saying "Client X disconnected"
                # for client in tclients:
                #     tserver.send_message(client, "Client {} disconnected.".format(disconnected_client))

            # For each message a client has sent
            for sender_client, message in tserver.get_messages():
                if sender_client not in tclients:
                    continue

                # Send every client a message reading: "I received "[MESSAGE]" from client [ID OF THE SENDER CLIENT]"
                # for client in tclients:
                #     tserver.send_message(client, 'I received "{}" from client {}'.format(message, sender_client))

            # process FT8, FT4, WSPR, FST4W, JS8 websocket messages from SparkSDR
            while not wsclient.responseQ.empty():
                # process FT8, FT4, WSPR, FST4W, JS8 websocket messages from SparkSDR queue
                spark_msgs = wsclient.responseQ.get()

                json_msgs = json.loads(spark_msgs)

                printed = False

                # for every individual decode spot
                if "spots" in json_msgs:
                    for json_msg in json_msgs["spots"]:
                        valid, hhmmss, freq, mode, call, grid, tone, snr, dt, dbm, msg, jtmsg = parse_message(json_msg)

                        # don't process or report call if it was recently seen within expire timeout
                        if valid == "FRESH" or call == "" or call == "TNX":
                            continue

                        if debug and (valid is False or valid == "false"):
                            print()
                            print("<<<--- INVALID --->>> {:6}  {:7.1f} {:>5} {:>11} {:>4} {:>3} {:>4}  {}".format(
                                hhmmss, freq, mode, call, tone, snr, grid, jtmsg))
                            print()
                            continue

                        dxline = "DX de {:<9}{:>9}  {:<13}{:>5}  {:<23} {:.4}Z". \
                            format(mycall+"-#:", freq, call, mode, msg, hhmmss)

                        if debug:
                            print(dxline)
                            printed = True

                        # Send every client the spot message
                        for client in tclients:
                            tserver.send_message(client, dxline)

                    if debug and printed:
                        print()

            time.sleep(1)

            now_time = int(time.time())

            # run database expire clean up once a minute
            if now_time >= clean_timer + (1 * 60):
                clean_calls_db()
                clean_timer = int(time.time())

    except KeyboardInterrupt:
        if debug:
            print("Closing main thread")

        p = psutil.Process(pid)
        p.terminate()  # or p.kill()

        wsclient.controller.join()
